
# -*- coding: utf-8 -*-
"""
Created on Wed Nov  8 14:41:57 2023

@autor: Prof. Heitor S. Lopes

ALGORITMO GENÉTICO PARA RESOLVER UMA INSTÂNCIA DO
CAIXEIRO VIAJANTE COM "C" CIDADES A VISITAR,
DISPOSTAS NUM ESPAÇO DE "ZxZ"

INFORMAÇÕES:
-AS COORDENADAS DAS CIDADES SÃO IMPORTADAS DE UM ARQUIVO CSV
-APÓS AS GERAÇÕES G/2 E G/2+G/4 A TAXA DE MUTAÇÃO É DOBRADA
-A MELHOR ROTA ENCONTRADA É MOSTRADA NUM GRÁFICO
-A EVOLUÇÃO DO FITNESS AO LONGO DAS GERAÇÕES É MOSTRADA NUM GRÁFICO
-NO FINAL É MOSTRADO O TEMPO DE PROCESSAMENTO
"""


import random
import math
import csv
import matplotlib.pyplot as plt
import time

start_time = time.time()

#CALCULA A DISTÂNCIA EUCLIDIANA ENTRE DUAS CIDADES
def euclidean_distance(city1, city2):
    x1, y1 = city1
    x2, y2 = city2
    return math.sqrt((x1 - x2)**2 + (y1 - y2)**2)

#CALCULA A DISTÂNCIA TOTAL PARA UMA ROTA QUE PASSA POR TODAS AS CIDADES
def total_distance(route, cities):
    distance = 0
    for i in range(len(route) - 1):
        distance += euclidean_distance(cities[route[i]], cities[route[i + 1]])
    return distance

#LÊ O ARQUIVO CSV COM AS COORDENADAS DAS CIDADES
def load_cities_from_csv(filename):
    cities = []
    with open(filename, 'r') as file:
        csv_reader = csv.reader(file)
        for row in csv_reader:
            city_name, x, y = row
            cities.append((float(x), float(y)))
    return cities

#GERA A POPULAÇÃO INICIAL ALEATORIAMENTE
def generate_initial_population(population_size, num_cities):
    population = [list(range(num_cities)) for _ in range(population_size)]
    for i in range(population_size):
        random.shuffle(population[i])
    return population

#EXECUTA O OPERADOR DE MUTAÇÃO COM UMA DETERMINADA PROBABILIDADE
def mutate(route, mutation_rate):
    if random.random() < mutation_rate:
        i, j = random.sample(range(len(route)), 2)
        route[i], route[j] = route[j], route[i]

#EXECUTA O OPERADOR DE CROSSOVER COM UMA DETERMINADA PROBABILIDADE
def crossover(parent1, parent2, crossover_rate):
    if random.random() < crossover_rate:
        n = len(parent1)
        start, end = sorted(random.sample(range(n), 2))
        child = [None] * n
        for i in range(start, end + 1):
            child[i] = parent1[i]
        remaining_cities = [city for city in parent2 if city not in child]
        j = 0
        for i in range(n):
            if child[i] is None:
                child[i] = remaining_cities[j]
                j += 1
        return child
    else:
        return parent1



#IMPLEMENTA AS ETAPAS DO ALGORITMO GENÉTICO
def genetic_algorithm(P, G, M, crossover_rate, cities):
    population = generate_initial_population(P, len(cities))
    best_fitness_history = []

    for generation in range(G):
        if (generation == G/2): 
            M=M*2  #Duplica a taxa de mutação a partir da metade das gerações
            print("Aumentando a mutação para", M,"na geração",generation)
        if (generation == int(G/2+G/4)): 
            M=M*2 #Duplica novamente a taxa de mutação a partir 3/4 das gerações
            print("Aumentando a mutação para", M, "na geração",generation)
        population = sorted(population, key=lambda x: total_distance(x, cities))
        best_route = population[0]
        best_fitness = total_distance(best_route, cities)
        best_fitness_history.append(best_fitness)

        new_population = population[:P // 2]

        for _ in range(P - P // 2):
            parent1, parent2 = random.sample(population[:P // 2], 2)
            child = crossover(parent1, parent2, crossover_rate)
            mutate(child, M)
            new_population.append(child)

        population = new_population

    best_route = min(population, key=lambda x: total_distance(x, cities))
    best_distance = total_distance(best_route, cities)
    return best_route, best_distance, best_fitness_history



#IMPRIME GRÁFICO DA MELHOR ROTA ENCONTRADA
def plot_tsp_solution(solution, cities):
    x = [cities[i][0] for i in solution]
    y = [cities[i][1] for i in solution]

    # Add the starting city at the end to complete the loop
    x.append(x[0])
    y.append(y[0])

    plt.figure(figsize=(8, 8))
    plt.plot(x, y, marker='o', linestyle='-')
    plt.scatter(x, y, c='red', marker='o')
    plt.title('Melhor solução - Problema do Caixeiro Viajante')
    plt.xlabel('Coordenada X')
    plt.ylabel('Coordenada Y')
    plt.grid(True)
    plt.show()



#IMPRIME O GRÁFICO DO FITNESS AO LONGO DAS GERAÇÕES
def plot_fitness_history(fitness_history):
    plt.figure(figsize=(8, 6))
    plt.plot(range(1, len(fitness_history) + 1), fitness_history, marker='o', linestyle='-')
    plt.title('Fitness por Gerações')
    plt.xlabel('Geração')
    plt.ylabel('Fitness (Distância total)')
    plt.grid(True)
    plt.show()




#PROGRAMA PRINCIPAL
if __name__ == "__main__":
    csv_filename = input("Arquivo CSV com as coordenadas: ")
    P = int(input("Tamanho da população: "))
    G = int(input("Número de gerações: "))
    M = float(input("Taxa de mutação: "))
    crossover_rate = float(input("Taxa de crossover (0..1): "))
    print("Rodando...")


    cities = load_cities_from_csv(csv_filename)

    best_route, best_distance, fitness_history = genetic_algorithm(P, G, M, crossover_rate, cities)
    print(f"Melhor rota: {best_route}")
    print(f"Distância total: {best_distance}")
    plot_tsp_solution(best_route, cities)
    plot_fitness_history(fitness_history)
    print("--- Tempo: %s segundos ---" % (time.time() - start_time))